@@ -1,23 +1,26 @@
 use 5.010001;
 use strict;
 use warnings;
+no warnings 'recursion';
 
 package BSON::PP;
 # ABSTRACT: Pure Perl BSON implementation
 
 use version;
-our $VERSION = 'v1.4.0';
+our $VERSION = 'v1.10.2';
 
 use B;
 use Carp;
 use Config;
 use Scalar::Util qw/blessed looks_like_number refaddr reftype/;
+use List::Util qw/first/;
 use Tie::IxHash;
 
 use BSON::Types ();
 use boolean;
+use mro;
 
-use if $] ge '5.010000', 're', 'regexp_pattern';
+use re 'regexp_pattern';
 
 use constant {
     HAS_INT64 => $Config{use64bitint},
@@ -49,6 +52,7 @@ use constant {
     BSON_REGEX => "Z*Z*",
     BSON_JSCODE => "",
     BSON_INT32 => "l<",
+    BSON_UINT32 => "L<",
     BSON_INT64 => "q<",
     BSON_8BYTES => "a8",
     BSON_16BYTES => "a16",
@@ -59,6 +63,7 @@ use constant {
     BSON_OBJECTID => 'a12',
     BSON_BINARY_TYPE => 'C',
     BSON_CSTRING => 'Z*',
+    BSON_MAX_DEPTH => 100,
 };
 
 sub _printable {
@@ -91,6 +96,20 @@ sub _ixhash_iterator {
     }
 }
 
+# relying on Perl's each() is prone to action-at-a-distance effects we
+# want to avoid, so we construct our own iterator for hashes
+sub _hashlike_iterator {
+    my $hashlike = shift;
+    my @keys = keys %$hashlike;
+    @keys = sort @keys
+        if $ENV{BSON_TEST_SORT_HASH};
+    return sub {
+        my $k = shift @keys;
+        return unless defined $k;
+        return ($k, $hashlike->{$k});
+    }
+}
+
 # XXX could be optimized down to only one substr to trim/pad
 sub _bigint_to_int64 {
     my $bigint = shift;
@@ -167,12 +186,25 @@ sub _encode_bson {
     my $refaddr = refaddr($doc);
     die "circular reference detected" if $opt->{_circular}{$refaddr}++;
 
+    $opt->{_depth} = 0 unless defined $opt->{_depth};
+    $opt->{_depth}++;
+    if ($opt->{_depth} > BSON_MAX_DEPTH) {
+        croak "Exceeded max object depth of ". BSON_MAX_DEPTH;
+    }
+
     my $doc_type = ref($doc);
 
-    return $doc->bson
-      if $doc_type eq 'BSON::Raw' || $doc_type eq 'MongoDB::BSON::_EncodedDoc';
+    if ( $doc_type eq 'BSON::Raw' || $doc_type eq 'MongoDB::BSON::_EncodedDoc' ) {
+        delete $opt->{_circular}{$refaddr};
+        $opt->{_depth}--;
+        return $doc->bson;
+    }
 
-    return $$doc if $doc_type eq 'MongoDB::BSON::Raw';
+    if ( $doc_type eq 'MongoDB::BSON::Raw' ) {
+        delete $opt->{_circular}{$refaddr};
+        $opt->{_depth}--;
+        return $$doc;
+    }
 
     my $iter =
         $doc_type eq 'HASH'           ? undef
@@ -182,6 +214,8 @@ sub _encode_bson {
       : $doc_type eq 'MongoDB::DBRef' ? _ixhash_iterator( $doc->_ordered )
       :                                 do { _reftype_check($doc); undef };
 
+    $iter //= _hashlike_iterator($doc);
+
     my $op_char = defined($opt->{op_char}) ? $opt->{op_char} : '';
     my $invalid =
       length( $opt->{invalid_chars} ) ? qr/[\Q$opt->{invalid_chars}\E]/ : undef;
@@ -192,7 +226,7 @@ sub _encode_bson {
     my $bson = '';
 
     my ($key, $value);
-    while ( $first_key_pending or ( $key, $value ) = $iter ? $iter->() : (each %$doc) ) {
+    while ( $first_key_pending or ( $key, $value ) = $iter->() ) {
         next if defined $first_key && $key eq $first_key;
 
         if ( $first_key_pending ) {
@@ -222,6 +256,14 @@ sub _encode_bson {
         utf8::encode($utf8_key);
         my $type = ref $value;
 
+        # If the type is a subtype of BSON::*, use that instead
+        if ( blessed $value ) {
+            if ($type !~ /\ABSON::\w+\z/) {
+                my $parent = first { /\ABSON::\w+\z/ } reverse @{mro::get_linear_isa($type)};
+                $type = $parent if defined $parent;
+            }
+        }
+
         # Null
         if ( !defined $value ) {
             $bson .= pack( BSON_TYPE_NAME, 0x0A, $utf8_key );
@@ -238,6 +280,16 @@ sub _encode_bson {
                 $bson .= pack( BSON_TYPE_NAME, 0x04, $utf8_key ) . _encode_bson( \%h, $opt );
             }
 
+            # special-cased deprecated DBPointer
+            elsif ($type eq 'BSON::DBPointer') {
+                my %data;
+                tie %data, 'Tie::IxHash';
+                $data{'$ref'} = $value->{'ref'};
+                $data{'$id'} = $value->{id};
+                $bson .= pack( BSON_TYPE_NAME, 0x03, $utf8_key )
+                    . _encode_bson(\%data, $opt);
+            }
+
             # Document
             elsif ($type eq 'HASH'
                 || $type eq 'BSON::Doc'
@@ -357,7 +409,7 @@ sub _encode_bson {
             }
 
             # String (explicit)
-            elsif ( $type eq 'BSON::String' ) {
+            elsif ( $type eq 'BSON::String' || $type eq 'BSON::Symbol') {
                 $value = $value->value;
                 utf8::encode($value);
                 $bson .= pack( BSON_TYPE_NAME.BSON_STRING, 0x02, $utf8_key, $value );
@@ -403,49 +455,21 @@ sub _encode_bson {
             }
         }
 
-        ### SCALAR: numbers preferred
-
-        elsif ( $opt->{prefer_numeric} ) {
-
-            # Int heuristic based on size
-            if ( $value =~ $int_re ) {
-                if ( $value > $max_int64 || $value < $min_int64 ) {
-                    croak("BSON can only handle 8-byte integers. Key '$key' is '$value'");
-                }
-                elsif ( $value > $max_int32 || $value < $min_int32 ) {
-                    $bson .= pack( BSON_TYPE_NAME, 0x12, $utf8_key ) . _pack_int64($value);
-                }
-                else {
-                    $bson .= pack( BSON_TYPE_NAME . BSON_INT32, 0x10, $utf8_key, $value );
-                }
-            }
-
-            # Double
-            elsif ( $value =~ $doub_re ) {
-                $bson .= pack( BSON_TYPE_NAME.BSON_DOUBLE, 0x01, $utf8_key, $value );
-            }
-
-            # String
-            else {
-                utf8::encode($value);
-                $bson .= pack( BSON_TYPE_NAME.BSON_STRING, 0x02, $utf8_key, $value );
-            }
-
-        }
-
-        ### SCALAR: strings preferred
-
+        # SCALAR
         else {
+            # If a numeric value exists based on internal flags, use it;
+            # otherwise, if prefer_numeric is true and it looks like a
+            # number, then coerce to a number of the right type;
+            # otherwise, leave it as a string
 
             my $flags = B::svref_2object(\$value)->FLAGS;
-            if ( $flags & B::SVf_POK() ) {
-                utf8::encode($value);
-                $bson .= pack( BSON_TYPE_NAME.BSON_STRING, 0x02, $utf8_key, $value );
-            }
-            elsif ( $flags & B::SVf_NOK() ) {
+
+            if ( $flags & B::SVf_NOK() ) {
                 $bson .= pack( BSON_TYPE_NAME.BSON_DOUBLE, 0x01, $utf8_key, $value );
             }
             elsif ( $flags & B::SVf_IOK() ) {
+                # Force numeric; fixes dual-vars comparison bug on old Win32s
+                $value = 0+$value;
                 if ( $value > $max_int64 || $value < $min_int64 ) {
                     croak("BSON can only handle 8-byte integers. Key '$key' is '$value'");
                 }
@@ -456,27 +480,42 @@ sub _encode_bson {
                     $bson .= pack( BSON_TYPE_NAME . BSON_INT32, 0x10, $utf8_key, $value );
                 }
             }
-            elsif ( $] lt '5.010' && $flags & B::SVp_IOK() ) {
-                if ( $value > $max_int64 || $value < $min_int64 ) {
-                    croak("BSON can only handle 8-byte integers. Key '$key' is '$value'");
+            elsif ( $opt->{prefer_numeric} && looks_like_number($value) ) {
+                # Looks like int: type heuristic based on size
+                if ( $value =~ $int_re ) {
+                    if ( $value > $max_int64 || $value < $min_int64 ) {
+                        croak("BSON can only handle 8-byte integers. Key '$key' is '$value'");
+                    }
+                    elsif ( $value > $max_int32 || $value < $min_int32 ) {
+                        $bson .= pack( BSON_TYPE_NAME, 0x12, $utf8_key ) . _pack_int64($value);
+                    }
+                    else {
+                        $bson .= pack( BSON_TYPE_NAME . BSON_INT32, 0x10, $utf8_key, $value );
+                    }
                 }
-                elsif ( $value > $max_int32 || $value < $min_int32 ) {
-                    $bson .= pack( BSON_TYPE_NAME, 0x12, $utf8_key ) . _pack_int64($value);
+
+                # Looks like double
+                elsif ( $value =~ $doub_re ) {
+                    $bson .= pack( BSON_TYPE_NAME.BSON_DOUBLE, 0x01, $utf8_key, $value );
                 }
+
+                # looks_like_number true, but doesn't match int/double
+                # regexes, so as a last resort we leave as string
                 else {
-                    $bson .= pack( BSON_TYPE_NAME . BSON_INT32, 0x10, $utf8_key, $value );
+                    utf8::encode($value);
+                    $bson .= pack( BSON_TYPE_NAME.BSON_STRING, 0x02, $utf8_key, $value );
                 }
             }
             else {
-                # Last resort, string
+                # Not coercing or didn't look like a number
                 utf8::encode($value);
                 $bson .= pack( BSON_TYPE_NAME.BSON_STRING, 0x02, $utf8_key, $value );
             }
-
         }
     }
 
     delete $opt->{_circular}{$refaddr};
+    $opt->{_depth}--;
 
     return pack( BSON_INT32, length($bson) + 5 ) . $bson . "\0";
 }
@@ -505,8 +544,8 @@ my %FIELD_SIZES = (
     0xFF => 0,
 );
 
-my $ERR_UNSUPPORTED = "Unsupported BSON type 0x%x for key '%s'.  Are you using the latest driver version?";
-my $ERR_TRUNCATED = "Premature end of BSON field '%s' (type 0x%x)";
+my $ERR_UNSUPPORTED = "unsupported BSON type \\x%X for key '%s'.  Are you using the latest version of BSON.pm?";
+my $ERR_TRUNCATED = "premature end of BSON field '%s' (type 0x%x)";
 my $ERR_LENGTH = "BSON field '%s' (type 0x%x) has invalid length: wanted %d, got %d";
 my $ERR_MISSING_NULL = "BSON field '%s' (type 0x%x) missing null terminator";
 my $ERR_BAD_UTF8 = "BSON field '%s' (type 0x%x) contains invalid UTF-8";
@@ -524,6 +563,11 @@ sub _decode_bson {
     if ( !defined $bson ) {
         croak("Decode argument must not be undef");
     }
+    $opt->{_depth} = 0 unless defined $opt->{_depth};
+    $opt->{_depth}++;
+    if ($opt->{_depth} > BSON_MAX_DEPTH) {
+        croak "Exceeded max object depth of ". BSON_MAX_DEPTH;
+    }
     my $blen= length($bson);
     my $len = unpack( BSON_INT32, $bson );
     if ( length($bson) != $len ) {
@@ -747,7 +791,7 @@ sub _decode_bson {
 
         # Timestamp
         elsif ( $type == 0x11 ) {
-            ( my $sec, my $inc, $bson ) = unpack( BSON_INT32.BSON_INT32.BSON_REMAINING, $bson );
+            ( my $sec, my $inc, $bson ) = unpack( BSON_UINT32.BSON_UINT32.BSON_REMAINING, $bson );
             $value = BSON::Timestamp->new( $inc, $sec );
         }
 
@@ -781,7 +825,9 @@ sub _decode_bson {
 
         # ???
         else {
-            croak "Unsupported type $type";
+            # Should have already been caught in the minimum length check,
+            # but just in case not:
+            croak( sprintf( $ERR_UNSUPPORTED, $type, $key ) );
         }
 
         if ( $opt->{_decode_array} ) {
@@ -791,6 +837,7 @@ sub _decode_bson {
             $hash{$key} = $value;
         }
     }
+    $opt->{_depth}--;
     return $opt->{_decode_array} ? \@array : \%hash;
 }
 
@@ -806,7 +853,7 @@ BSON::PP - Pure Perl BSON implementation
 
 =head1 VERSION
 
-version v1.4.0
+version v1.10.2
 
 =head1 DESCRIPTION
 
@@ -830,7 +877,7 @@ Stefan G. <minimalist@lavabit.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2017 by Stefan G. and MongoDB, Inc.
+This software is Copyright (c) 2018 by Stefan G. and MongoDB, Inc.
 
 This is free software, licensed under:
 
